# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import sys

from fedrq.cli import main

main(prog=f"{sys.executable} -m fedrq")
